# Présentation eBPF

## Qu'est ce que c'est ?

### Histoire

Vient de BPF (Berkeley Packet Filter), petite machine virtuelle à l'intérieur du kernel Linux pour faire des règles de filtrages directement dans le kernel, première apparition en 1993, intégrée au noyau Linux en 1997 avec la version 2.1.75.

Donc au début, c'était surtout pour filtrer des paquets réseaux depuis le kernelspace plutôt que tout réclamer puis filtrer depuis le userspace, d'où le nom (tcpdump est basé dessus depuis longtemps).

2012 -> évolution pour sécuriser et filtrer les appels systèmes

Arrivée de eBPF (extended BPF) en 2013 : extension des capacités de la VM. ET SURTOUT :
Un vrai langage avec un vérifieur pour exécuter du code dans le kernel (C très bordé : pas de pointeurs, pas de boucles infinies, includes limités...), et des fonctions spécifiques

Depuis 2013 il y a des ajouts de fonctionnalités dans chaque nouvelles versions du kernel

Donc c'est plus seulement un outil uniquement pour le réseau, désormais on peut ajouter du tracing sur des appels systèmes (utilise des probes, ou tracepoints, qui existent déjà depuis longtemps) ou des évènements du kernel, tracer des programmes grâce a leurs propres probes, etc. C'est là que ça devient très intéressant

Le code eBPF ne peut pas faire crasher le système parce qu'il est utilisé en sandbox dans le kernel, avec un nombre d'instructions et des ressources limitées. C'est du bytecode, en Just In Time

### Resumé

C'est une petite VM dans le kernel, dans laquelle on injecte du code C très bordé qui utilise des fonctions données par eBPF afin de récupérer des infos précises sur ce qu'on cherche

## Utilisation

Il y a un outil qui est fourni, qui est la manière "simple" d'utiliser eBPF -> BCC (BPF Compiler Collection - https://github.com/iovisor/bcc)

Supporté intégralement depuis la version 4.9 du kernel (décembre 2016)

BCC propose un frontend python (ou lua, go) pour écrire des programmes qui vont injecter du C dans le kernel et récupérer les infos en python pour les formater, les traiter...

Exemple très simple, on branche un event dans le kernel qui va nous envoyer un message "Hello, eBPF!" à chaque fois qu'un processus est créé

En utilisant les fonctions **trace**, le kernel va écrire sa réponse dans le fichier "/sys/kernel/debug/tracing/trace_pipe", que l'on va récupérer dans notre code python. C'est la manière la plus simple de faire, mais il existe d'autres manières pour des besoins plus spécifiques qu'afficher un message...
Par exemple, créer une structure dans le programme eBPF (en C donc) et de créer une structure équivalente en python (grâce au module ctypes) et refléter les données de la structure C sur cette dernière.

```python
#!/usr/bin/env python

import os
from bcc import BPF

# Le programme que l'on va injecter dans le kernel
program='''
int kprobe__sys_clone(void *ctx) {
  bpf_trace_printk("Hello, eBPF!\\n");
  return 0;
}
'''

# On lance notre programme
b = BPF(text = program)

while 1:
	# Récuperation des informations envoyées par le kernel
    (task, pid, cpu, flags, ts, msg) = b.trace_fields()
	# Formatage des infos
    print("PID = %u\nPROGRAM = %s\nMESSAGE = %s\n" % (pid, task, msg))
```

BCC fournit aussi beaucoup de scripts et tools déjà faits pour diverses tâches

![bcc_tracing_tools](bcc_tracing_tools.png)

Quelques exemples :

- vfsstat : Nombre de syscalls read, write, open...
- bashreadline : Récupère toutes les commandes bash lancées avec le pid, user et commande

Des trucs plus fun:

Pour observer du code PHP, il faut que php soit compilé avec l'option de debug qui active les tracepoints `./configure --enable-dtrace`

Voir toutes les fonctions PHP appelées dans l'ordre, avec Classe.Method -> fichier
- `sudo ./trace -p PID 'u::function__entry "%s.%s -> %s" arg4,arg1,arg2' -T`

Voir le flot d'appel de fonctions PHP dans l'ordre
- `sudo ./uflow -l php PID`

Voir le nombre d'appel pour chaque fonction PHP
- `sudo ./ucalls -L -l php PID`

Voir le détail de certains appels systèmes faits par un programme. Par exemple les opens, le fichier ouvert et le temps
- `sudo ./trace -p PID 'do_sys_open "%s", arg2' -t`

BCC fournit aussi un outil en ligne de commande pour faire la même chose : bpftrace. Peut être utile pour debugger à chaud

Le fait de pouvoir exécuter du code en mode kernel est très intéressant, pas besoin de recompiler le kernel ou de développer des modules
