# Titre Article
# eBPF Linux superpowers?

## eBP... quoi?

eBPF est un outil de tracing du noyau Linux qui permet d'instrumenter le kernel afin de récupérer des evenements et de pouvoir agir en consequence.

### Un peu d'histoire

L'idée est née avec BPF (Berkeley Package Filter), apparu pour la première fois en 1993 puis intégré au kernel en 1997 avec la version 2.1.75. BPF était une petite machine virtuelle à l'intérieur du kernel où on pouvait y injecter des instructions pour faire des règles de filtrage réseau.

L'usage de base était donc uniquement de filtrer des paquets réseau directement depuis le kernelspace plutôt que tout réclamer puis filtrer depuis le userspace, d'où le nom "Packet Filter". Tcpdump est basé dessus depuis longtemps.

Une première évolution est arrivée en 2012 avec la gestion des appels systèmes, puis en 2013 sort enfin eBPF, ou Extended Berkeley Package Filter, qui est maintenant amélioré à chaque nouvelle release.

Avec eBPF arrive une extension des capacités de la vm et surtout une vraie interface avec celle ci. On va pouvoir y injecter du code C à executer. Mais attention, il ne s'agit pas d'aller y faire n'importe quoi, il y a un vérifieur qui va controler le programme injecté : pas de boucles infinies, pas de pointeurs, pas d'allocation dynamique... eBPF fournit aussi beaucoup de fonctions pour notre programme. Vous l'aurez compris, le programme injecté sera très bordé. De plus, étant dans une vm, le programme est executé en sandbox, avec un nombre d'instructions et de ressources limités, et compilé en JIT. La possibilité de faire crasher le kernel est ainsi éliminée.

Ce n'est donc plus seulement un outil pour le filtrage réseau mais un réel outil de tracing à l'intérieur même du kernel. On pourra tracer les appels systèmes (en utilisant les kprobes ou tracepoints qui existent déjà depuis bien longtemps), tracer les langages qui fournissent leurs propres probes (comme nous le verrons un peu plus loin), et plus généralement avoir une visibilité sur à peut près tout ce qui passe par le kernel...

![gif](https://media.giphy.com/media/d3mlE7uhX8KFgEmY/giphy.gif)

### Utilisation

Il y a plusieurs manière d'intéragir avec eBPF, en C via notamment l'utilisation de l'appel système `bpf`, ou en utilisant un "frontend" qui va nous fournir une interface dans un langage de plus haut niveau, ou encore l'outil en ligne de commande `bpftrace`.

Bpftrace est un outil en ligne de commande principalement fait pour récupérer des informations et debbuger à chaud. (https://github.com/iovisor/bpftrace).

Nous allons surtout nous interesser à BCC, un frontend python (mais aussi lua, golang...) maintenu par le projet open source IO Visor (https://www.iovisor.org/) de la Linux Foundation (https://www.linuxfoundation.org/projects/).

BCC est supporté intégralement depuis la version 4.9 du kernel (décembre 2016), c'est un outil relativement récent et encore en développement. Il se trouve dans les paquets debian sous sa version 0.8.0.

Le projet fournit beaucoup de scripts et tools généralistes déjà faits pour diverses taches (paquet debian `bpfcc-tools`), surveiller l'appel a certains syscalls, monitorer l'utilisation du reseau, détecter les fuites de memoire, etc. Ces outils sont intégrés par defaut sur les serveurs chez Netflix et Facebook, entre autres...
Voici la liste officielle:

![bcc_tracing_tools](assets/bcc_tracing_tools.png)

Vous trouverez des informations sur chacun d'entre eux sur le github du projet dans le dossier "tools" (https://github.com/iovisor/bcc/tree/master/tools).

Interessons nous maintenant à une partie plus excitante : l'écriture d'outils avec BCC. Sous debian, nous avons besoin des paquets `bpfcc-tools` et `bpfcc-python`, ou `bpfcc-python3` en fonction de la version de python utilisée. Voici un exemple simple :

```python
#!/usr/bin/env python

import os
from bcc import BPF

# Le programme C que l'on va injecter dans le kernel
program='''
int kprobe__sys_clone(void *ctx) {
  bpf_trace_printk("Hello World\\n");
  return 0;
}
'''

# On lance notre programme
b = BPF(text = program)

while 1:
	# Récuperation des informations envoyées par le kernel
    (task, pid, cpu, flags, ts, msg) = b.trace_fields()
	# Formatage des infos
    print("PID = %u\nPROGRAM = %s\nMESSAGE = %s\n" % (pid, task, msg))
```

Ce programme injecte un code C dans le kernel, qui va nous renvoyer un message "Hello World" à chaque fois qu'un processus est créé. Cela se décompose en 3 étapes : la création du programme C, son injection dans eBPF et la récupération et traitement des informations en python.

Dans ce programme d'exemple, nous utilisons la technique simple du fichier /sys/kernel/debug/tracing/trace_pipe. eBPF écrit dedans ce qu'on lui à demandé avec la fonction `bpf_trace_printk`, et nous récupérons son contenu avec la fonction `trace_fields` dans notre code python. Cette méthode est bien pour un programme simple mais n'est pas recommandée.

Il existent deux autres méthodes bien plus efficace : les structures MAPS et le tampon PERF.
Les structures MAPS sont un type de données propre à eBPF, un simple tableau clés/valeurs permettant de stocker des données, qui seront gardées à chaque nouvelle éxecution du programme. On pourra ainsi stocker des informations persistantes comme des statistiques sur le nombre d'appels systèmes, des compteurs etc.
Le tampon PERF est une technique qui permet de stocker des informations dans une structure C classique, puis de la récupérer et la parser dans une structure python et ainsi accéder à tous ses champs directement dans notre script. Les structures C disparaitront à chaque fin d'éxecution du programme eBPF. On comprends donc qu'il faudra utiliser les deux techniques afin de renvoyer des structures avec les informations et de stocker nos compteurs ou statistiques dans des MAPS.

Un programme eBPF va surveiller des "tracepoints", il en existe 4 types : **tracepoint**, **kprobe**, **usdt** et **uprobe**. Les **kprobe** et **uprobe** sont des tracepoints "dynamiques", c'est à dire qu'ils vont récupérer des informations à chaud sur les programmes utilisateurs pour les **uprobe** et kernel pour les **kprobe**. Ils ne necessitent pas la recompilation du programme mais sont susceptibles de changer en fonction des versions du kernel.
Quant aux **tracepoint** et **usdt**, ce sont des tracepoints "statiques", ils sont définis directement dans le code source du programme, utilisateur pour **usdt** et kernel pour **tracepoint**. La re-compilation du programme est donc necessaire.

Différents tracepoints sont donc exposés par les programmes, et nous nous branchons dessus pour récupérer leurs évènements grace à eBPF.

Nous venons de voir que les tracepoints statiques appartiennent aux programmes, ils sont donc accessibles par eBPF via le PID du processus de ces programmes. Quelque chose devrait commencer à vous venir en tête...

Si nous regardons la liste des processus sur un serveur, qui contient des programmes qui tournent, et des conteneurs... Nous voyons aussi les processus lancés dans ces fameux conteneurs.

![screenshot](assets/screenshot.png)

Pour ceux que ça étonne, un petit rappel sur ce qu'est un conteneur est necessaire (https://itnext.io/chroot-cgroups-and-namespaces-an-overview-37124d995e3d?gi=75bd68815b75)

Cela veut donc dire qu'avec les tracepoints, nous allons pouvoir instrumenter n'importe quel programme compatible, qu'il soit dans un conteneur ou non.

### Présentation du tool

Nous avons crée un programme BCC permettant de tracer et comprendre le comportement d'une application PHP. Le code est disponible sur github (https://github.com/factorysh/PHP-tracing-tool).

Cet outil en ligne de commmande va instrumenter le code PHP du proccessus passé en paramètre grace aux tracepoints usdt du langage. PHP va donc devoir être compilé avec l'option `--enable-dtrace` pour les activer... Quelle chance, l'image PHP avec l'option de debug se trouve aussi dans le repository!

Il y a également deux démos d'applications PHP dans le projet afin de tester, une basique qui fait un appel à une API (google maps), ouvre un fichier et quitte. Et une seconde à peine plus grosse... (lol) Un Wordpress.
Nous allons tester le tool avec la demo simple.

Une fois les conteneurs lances, reprenons la liste des processus que nous avions un peu plus haut, choisissons en un PHP qui tourne dans notre conteneur (php-fpm en l'occurence) et lançons notre tool avec ce PID en paramètre et l'option `-S` qui active l'affichage du détail des appels systèmes.

![screenshot1](assets/screenshot1.png)

J'ai choisi d'activer uniquement les appels systèmes réseaux pour cet exemple mais il est possible de tracer tous les appels systèmes. Il suffit de les ajouter dans la liste déclarée en haut du code.

On peut donc voir, par exemple, que la fonction `geocodeAddress` qui appartient à la classe `GmapApi` s'est éxécutée en **4175791975ns**. A l'intérieur de celle ci se trouve les syscalls que l'on a choisi de suivre, avec leurs latences et des informations les concernants, comme l'adresse jointe par `connect` ou encoer le filedescriptor retourné par `socket`. On a aussi un résumé du temps passé dans les appels systemes que nous surveillons, du temps passé dans des appels systemes reseaux ou disque (il n'y en a pas ici), ainsi que le volume d'octets ecrits et lus.

Cet outil est une facon de decouvrir les possibilitees ouverte par eBPF et la facon dont on peut l'utiliser pour instrumenter les evenements d'un langage ou du kernel, en gardant une performance exceptionnelle etant donné que le code est execute directement dans le kernelspace et que les echanges kernelspace/userspace sont limites le plus possible.
Cela donne des outils performants, et utilisable en boite noire dans un environement reel, comme une production, pour comprendre le comportement de son application et ainsi detecter des fautes de programmation, des problemes de librairies ou d'API tierces...
